/*
 * Copyright (c) 1983 Eric P. Allman
 * Copyright (c) 1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted provided
 * that: (1) source distributions retain this entire copyright notice and
 * comment, and (2) distributions including binaries display the following
 * acknowledgement:  ``This product includes software developed by the
 * University of California, Berkeley and its contributors'' in the
 * documentation or other materials provided with the distribution and in
 * all advertising materials mentioning features or use of this software.
 * Neither the name of the University nor the names of its contributors may
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
#ifdef DBM
static char sccsid[] = "@(#)alias.c	5.21 (Berkeley) 6/1/90 (with DBM)";
#else
static char sccsid[] = "@(#)alias.c	5.21 (Berkeley) 6/1/90 (without DBM)";
#endif
#endif /* not lint */

# include <sys/types.h>
# include <sys/stat.h>
# include <signal.h>
# include <errno.h>
# include "sendmail.h"
# include <sys/file.h>
# include <pwd.h>

#ifdef X500
#include "dixie.h"
#include "x500.h"

char	*x500searchbase;
int	x500matchtype = 'E';
int	x500errno;
int	x500numerrs;
int	x500howmany;
char	**x500text;
char	*dixiehost;

static char *x500sendmailversion = "x500sendmail version = (5.64)1.1\n";

static char     dn[256];
static char     x500name[512];
static char     x500buf[4096];
char		x500answer[MAX_PKT_SIZE];

/*
**	X500ONEMATCH -- handle one match to an X.500 name
**
**	Parameters:
**		ent -- the buffer holding the match, as returned
**			by dxi_firstentry() or dxi_nextentry().
**
**	Returns:
**		the value of the name, if a single match.
**		NULL if unknown or multiple matches.
**
**	Side Effects:
**		May set x500numerrs, x500text if there was
**		no mail attribute for the entry.
**
**	Warnings:
**		The return value will be trashed across calls.
*/

char *x500onematch( ent )
char	*ent;
{
	char	**mailav, **addrav, **phoneav, **titleav;
	int	len, i;
	char	*result, *t, *p, *dn, *rdn;

	mailav = dxi_getvalues( ent, "mail" );

	/* there are one or more rfc822mail attributes */
	if ( mailav != (char **) 0 ) {
		len = 0;
		for ( i = 0; mailav[i] != NULL; i++ )
			len += strlen( mailav[i] ) + 2;

		result = (char *) malloc( len );
		strcpy( result, mailav[0] );
		for ( i = 1; mailav[i] != NULL; i++ ) {
			strcat( result, "," );
			strcat( result, mailav[i] );
		}

		dxi_valuefree( mailav );

		return( result );

	/* no mail attribute - bounce with address and phone */
	} else {
		addrav = dxi_getvalues( ent, "postaladdress" );
		phoneav = dxi_getvalues( ent, "telephonenumber" );
		titleav = dxi_getvalues( ent, "title" );
		dn = dxi_getdn( ent );
		rdn = dxi_lastdncomp( dn );

		/* fill in the error message */
		x500numerrs++;
		if ( x500text )
			x500text = (char **) realloc(x500text, 
			    sizeof(char *)*x500numerrs);
		else
			x500text = (char **)
			    malloc(sizeof(char *));
		if ( x500text == NULL ) {
			syslog(LOG_ALERT, "x500 alias malloc failed!");
			return( NULL );
		}

		if ( addrav != (char **) 0 && phoneav != (char **) 0 )
			sprintf(x500buf,"%cName, title, postal address and phone for '%s':\n\t", X_NOEMAIL, x500name);
		else if ( addrav != (char **) 0 )
			sprintf(x500buf,"%cName, title and postal address for '%s':\n\t", X_NOEMAIL, x500name);
		else if ( phoneav != (char **) 0 )
			sprintf(x500buf,"%cName, title and telephone number for '%s':\n\t", X_NOEMAIL, x500name);
		else
			sprintf(x500buf,"%c'%s' has no postal address or phone number registered in this service.\n\t", X_NOEMAIL, x500name);

		t = index(x500buf, '\0');
		/* print the official name */
		strcpy( t, rdn );
		strcat( t, "\n\t\t" );

		/* print the title, if available */
		if ( titleav != (char **) 0 ) {
			for ( i = 0; titleav[i] != NULL; i++ ) {
				strcat( t, titleav[i] );
				strcat( t, "\n\t\t" );
			}
		}

		t = index(t, '\0');
		/* print the address, if available */
		if ( addrav != (char **) 0 ) {
			p = addrav[0];
			while (*p) {
				switch (*p) {
				case '$':
					*t++ = '\n';
					*t++ = '\t';
					*t++ = '\t';
					break;
				default :
					*t++ = *p;
					break;
				}
				p++;
			}
			*t++ = '\n';
			*t++ = '\t';
			*t++ = '\t';
			*t = '\0';
		}

		/* print the phone number(s), if available */
		if ( phoneav != (char **) 0 ) {
			for ( i = 0; phoneav[i] != NULL; i++ ) {
				strcat( t, phoneav[i] );
				strcat( t, "\n\t\t" );
			}
		}

		dxi_valuefree( mailav );
		dxi_valuefree( phoneav );
		dxi_valuefree( addrav );
		dxi_valuefree( titleav );
		free( dn );

		x500text[x500numerrs-1] = newstr(x500buf);
		x500errno = X_NOEMAIL;
		return( NULL );
	}
	/* NOT REACHED */
}

/*
**	X500MATCH -- handle one or more matches to an X.500 name
**
**	Parameters:
**		buf -- the buffer holding the match(es), as returned
**			by dxi_search().
**		dncheck -- if true, give preference to an entry whose
**			dn contains x500name as the last component
**
**	Returns:
**		the value of the name, if a single match.
**		NULL if unknown or multiple matches.
**
**	Side Effects:
**		May set x500numerrs, x500text if there was
**		an approximate or multiple matching problem.
**
**	Warnings:
**		The return value will be trashed across calls.
*/

char *x500match( buf, dncheck )
char	*buf;
int	dncheck;
{
	char	*dn, *rdn, *ent;
	char	**mailav, **addrav, **phoneav, **titleav, **memberav;
	int	len, i, matches;
	char	*result, *p;

	/* a limit was exceeded */
	if ( dxi_errno == DXI_LIMIT ) {
		x500howmany = dxi_matches( buf );
		x500errno = X_TOOMANY;
		return( NULL );
	}

	if ( dxi_errno != DXI_SUCCESS ) {
		syslog(LOG_ALERT, "x500 error return from dixie (%d)",
		    dxi_errno);
		return( NULL );
	}

	/* one match */
	if ( (matches = dxi_matches( buf )) == 1 ) {
		ent = dxi_firstentry( buf );
		memberav = dxi_getvalues( ent, "member" );

		/* no members => not a group */
		if ( memberav == (char **) 0 )
			return( x500onematch( ent ) );

		/* 
		 * otherwise it's a group.  the plan is to return an
		 * alias that is all the member names (last parts only!)
		 * separated by commas.
		 */

		len = 0;
		for ( i = 0; memberav[i] != NULL; i++ )
			len += strlen( rindex( memberav[i], '=' ) ) + 2;

		result = (char *) malloc( len );
		rdn = dxi_lastdncomp( memberav[0] );
		strcpy( result, rdn );
		for ( i = 1; memberav[i] != NULL; i++ ) {
			rdn = dxi_lastdncomp( memberav[i] );
			strcat( result, "," );
			strcat( result, rdn );
		}
		for ( p = result; *p; p++ )
			if ( *p == ' ' )
				*p = '.';

		dxi_valuefree( memberav );

		return( result );

	/* more than one match - bounce with a list of possibilities */
	} else {
		/* first do dn check, if required */
		if ( dncheck ) {
			ent = dxi_firstentry( buf );
			for ( i = 0; i < matches; i++ ) {
				dn = dxi_getdn( ent );
				rdn = dxi_lastdncomp( dn );

				if ( strcasecmp( rdn, x500name ) == 0 ) {
					free( dn );
					return( x500onematch( ent ) );
				}

				free( dn );
				ent = dxi_nextentry();
			}
		}

		/* fill in the error message */
		x500howmany = matches;
		x500numerrs++;
		if ( x500text )
			x500text = (char **) realloc(x500text, 
			    sizeof(char *) * x500numerrs);
		else
			x500text = (char **) malloc(sizeof(char *));
		if ( x500text == NULL ) {
			syslog(LOG_ALERT, "x500 alias malloc failed!");
			return( NULL );
		}

		sprintf(x500buf,"%cThere are %d possible matches for '%s':\n\t", X_AMBIGUOUS, matches, x500name);

		ent = dxi_firstentry( buf );
		for ( i = 0; i < matches; i++ ) {
			char	**av;
			int	j;

			dn = dxi_getdn( ent );
			rdn = dxi_lastdncomp( dn );
			mailav = dxi_getvalues( ent, "mail" );
			phoneav = dxi_getvalues( ent, "telephonenumber" );
			titleav = dxi_getvalues( ent, "title" );
			av = mailav == (char **) 0 ? phoneav : mailav;

			strcat( x500buf, rdn );
			strcat( x500buf, "\t\t" );
			strcat( x500buf, mailav != (char **) 0 ? mailav[0]
			    : phoneav != (char **) 0 ? phoneav[0] : "" );
			strcat( x500buf, "\n\t" );
			if ( titleav != (char **) 0 ) {
				for ( j = 0; titleav[j] != NULL; j++ ) {
					strcat( x500buf, "\t" );
					strcat( x500buf, titleav[j] );
					strcat( x500buf, "\n\t" );
				}
			}

			dxi_valuefree( mailav );
			dxi_valuefree( phoneav );
			dxi_valuefree( titleav );
			free( dn );

			ent = dxi_nextentry();
		}

		x500text[x500numerrs-1] = newstr(x500buf);
		x500errno = X_AMBIGUOUS;
		return( NULL );
	}
}

/*
**	X500LOOKUP -- look up a name in the x.500 database
**
**	Parameters:
**		name -- the name to look up.
**
**	Returns:
**		the value of name.
**		NULL if unknown.
**
**	Side Effects:
**		May set x500errno if there was a temporary error.
**		May set x500numerrs, x500text if there was
**		an approximate or multiple matching problem.
**
**	Warnings:
**		The return value will be trashed across calls.
*/

static char *x500lookup(name)
char	*name;
{
	register char	*f, *p, *s, *tmp;
	register int	i;
	register char	*t;
	char		srchbuf[256];
	static char	*returnattrs[] = { "member", "mail", "title",
				"postaladdress", "telephonenumber", 0 };


	x500errno = 0;

	/* 
	 * don't bother looking if this person is already in the password
	 * file.  Otherwise, we might soundex match a loginid.
	 */

	if ( getpwnam(name) )
		return(NULL);

	/* 
	 * make the name cannonical - change '_' and '.' to ' '
	 * the canonicalized result is left in x500name, and the
	 * corresponding DN is left in dn.
	 */

	f = name;
	p = x500name;
	while (*f)
		switch (*f) {
		case '"':
			f++;
			break;

		case '.':	
		case '_':
			*p++ = ' ';
			f++;
			break;

		default	:
			*p++ = *f++;
		}
	*p = '\0';

	strcpy(dn, x500searchbase);
	strcat(dn, "@cn=");
	strcat(dn, x500name);

	if ( tTd(66,1) )
		dxi_debug = 1;

	/*
	 * First, search by uid.  There should only be zero or one uid
	 * matches, but we will bounce with choices if there are more.
	 * If we don't find that, search by commonName.  If more than
	 * one commonNames match, we give priority to any that are
	 * used to form the DN.  Finally, if cn does not match, we
	 * search by sn.  If one exact match is found, it is 
         * returned.  More than one exact match causes a bounce message
	 * to be generated.  If approximate (soundex) matching is wanted,
	 * x500matchtype should be changed from 'E' to 'S'.
	 */

	if ( dxi_init( dixiehost, 2, 4 ) == -1 ) {
		syslog(LOG_ALERT, "dxi_init failed!");
		return(NULL);
	}

	if ( x500matchtype == 'E' )
		sprintf( srchbuf, "uid=%s", x500name );
	else
		sprintf( srchbuf, "uid~=%s", x500name );

	if ( dxi_search( x500searchbase, DXI_SRCH_SUBTREE, srchbuf,
	    returnattrs, x500answer ) == -1 ) {
		/* queue it */
		x500errno = X_TRYAGAIN;
		return(NULL);
	}

	/* we found one or more uid matches -- handle it */
	if ( dxi_matches( x500answer ) > 0 )
		return( x500match( x500answer, 0 ) );

	if ( x500matchtype == 'E' )
		sprintf( srchbuf, "cn=%s", x500name );
	else
		sprintf( srchbuf, "cn~=%s", x500name );

	if ( dxi_search( x500searchbase, DXI_SRCH_SUBTREE, srchbuf,
	    returnattrs, x500answer ) == -1 ) {
		/* queue it */
		x500errno = X_TRYAGAIN;
		return(NULL);
	}

	/* we found one or more cn matches -- handle it */
	if ( dxi_matches( x500answer ) > 0 )
		return( x500match( x500answer, 1 ) );

	if ( x500matchtype == 'E' )
		sprintf( srchbuf, "sn=%s", x500name );
	else
		sprintf( srchbuf, "sn~=%s", x500name );

	if ( dxi_search( x500searchbase, DXI_SRCH_SUBTREE, srchbuf,
	    returnattrs, x500answer ) == -1 ) {
		/* queue it */
		x500errno = X_TRYAGAIN;
		return(NULL);
	}

	/* we found one or more sn matches -- handle it */
	if ( dxi_matches( x500answer ) > 0 )
		return( x500match( x500answer, 0 ) );

	return(NULL);
}
#endif /* X500 */

/*
**  ALIAS -- Compute aliases.
**
**	Scans the alias file for an alias for the given address.
**	If found, it arranges to deliver to the alias list instead.
**	Uses libdbm database if -DDBM.
**
**	Parameters:
**		a -- address to alias.
**		sendq -- a pointer to the head of the send queue
**			to put the aliases in.
**
**	Returns:
**		none
**
**	Side Effects:
**		Aliases found are expanded.
**
**	Notes:
**		If NoAlias (the "-n" flag) is set, no aliasing is
**			done.
**
**	Deficiencies:
**		It should complain about names that are aliased to
**			nothing.
*/


#ifdef DBM
typedef struct
{
	char	*dptr;
	int	dsize;
} DATUM;
extern DATUM fetch();
#endif DBM

alias(a, sendq)
	register ADDRESS *a;
	ADDRESS **sendq;
{
	register char *p;
	extern char *aliaslookup();

	if (tTd(27, 1))
		printf("alias(%s)\n", a->q_paddr);

	/* don't realias already aliased names */
	if (bitset(QDONTSEND, a->q_flags))
		return;

	CurEnv->e_to = a->q_paddr;

	/*
	**  Look up this name
	*/

	if (NoAlias) {
		p = NULL;
	} else {
		loweraddr( a );
		p = aliaslookup(a->q_user);
#ifdef X500
		if ( p == NULL )
			p = x500lookup( a->q_user );

		if ( p == NULL && x500errno == X_TRYAGAIN ) {
			markfailure( CurEnv, a, EX_TEMPFAIL );
			giveresponse( EX_TEMPFAIL, a->q_mailer, CurEnv );
			a->q_flags |= QNOTSURE;
			errno = 0;
		} else if ( p == NULL && x500errno == X_TOOMANY ) {
			markfailure( CurEnv, a, X_TOOMANY );
			giveresponse( X_TOOMANY, a->q_mailer, CurEnv );
			a->q_flags |= QDONTSEND;
			errno = 0;
		}
#endif
	}
	if (p == NULL)
		return;

#if defined(DBM) && defined(UUCPDOMAIN) || defined(X500)
	p = newstr( p );
#endif

	/*
	**  Match on Alias.
	**	Deliver to the target list.
	*/

	if (tTd(27, 1))
		printf("%s (%s, %s) aliased to %s\n",
		    a->q_paddr, a->q_host, a->q_user, p);
	message(Arpa_Info, "aliased to %s", p);
	AliasLevel++;
	sendtolist(p, a, sendq);
	AliasLevel--;
}
/*
**  ALIASLOOKUP -- look up a name in the alias file.
**
**	Parameters:
**		name -- the name to look up.
**
**	Returns:
**		the value of name.
**		NULL if unknown.
**
**	Side Effects:
**		none.
**
**	Warnings:
**		The return value will be trashed across calls.
*/

char *
aliaslookup(name)
	char *name;
{
# ifdef DBM
	DATUM rhs, lhs;

	/* create a key for fetch */
	lhs.dptr = name;
	lhs.dsize = strlen(name) + 1;
	rhs = fetch(lhs);
	return (rhs.dptr);
# else DBM
	register STAB *s;

	s = stab(name, ST_ALIAS, ST_FIND);
	if (s == NULL)
		return (NULL);
	return (s->s_alias);
# endif DBM
}
/*
**  INITALIASES -- initialize for aliasing
**
**	Very different depending on whether we are running DBM or not.
**
**	Parameters:
**		aliasfile -- location of aliases.
**		init -- if set and if DBM, initialize the DBM files.
**
**	Returns:
**		none.
**
**	Side Effects:
**		initializes aliases:
**		if DBM:  opens the database.
**		if ~DBM: reads the aliases into the symbol table.
*/

# define DBMMODE	0644

initaliases(aliasfile, init)
	char *aliasfile;
	bool init;
{
#ifdef DBM
	int atcnt;
	time_t modtime;
	bool automatic = FALSE;
	char buf[MAXNAME];
#endif DBM
	struct stat stb;
	static bool initialized = FALSE;

	if (initialized)
		return;
	initialized = TRUE;

	if (aliasfile == NULL || stat(aliasfile, &stb) < 0)
	{
		if (aliasfile != NULL && init)
			syserr("Cannot open %s", aliasfile);
		NoAlias = TRUE;
		errno = 0;
		return;
	}

# ifdef DBM
	/*
	**  Check to see that the alias file is complete.
	**	If not, we will assume that someone died, and it is up
	**	to us to rebuild it.
	*/

	if (!init)
		dbminit(aliasfile);
	atcnt = SafeAlias * 2;
	if (atcnt > 0)
	{
		while (!init && atcnt-- >= 0 && aliaslookup("@") == NULL)
		{
			/*
			**  Reinitialize alias file in case the new
			**  one is mv'ed in instead of cp'ed in.
			**
			**	Only works with new DBM -- old one will
			**	just consume file descriptors forever.
			**	If you have a dbmclose() it can be
			**	added before the sleep(30).
			*/

			sleep(30);
# ifdef NDBM
			dbminit(aliasfile);
# endif NDBM
		}
	}
	else
		atcnt = 1;

	/*
	**  See if the DBM version of the file is out of date with
	**  the text version.  If so, go into 'init' mode automatically.
	**	This only happens if our effective userid owns the DBM.
	**	Note the unpalatable hack to see if the stat succeeded.
	*/

	modtime = stb.st_mtime;
	(void) strcpy(buf, aliasfile);
	(void) strcat(buf, ".pag");
	stb.st_ino = 0;
	if (!init && (stat(buf, &stb) < 0 || stb.st_mtime < modtime || atcnt < 0))
	{
		errno = 0;
		if (AutoRebuild && stb.st_ino != 0 && stb.st_uid == geteuid())
		{
			init = TRUE;
			automatic = TRUE;
			message(Arpa_Info, "rebuilding alias database");
#ifdef LOG
			if (LogLevel >= 7)
				syslog(LOG_INFO, "rebuilding alias database");
#endif LOG
		}
		else
		{
#ifdef LOG
			if (LogLevel >= 7)
				syslog(LOG_INFO, "alias database out of date");
#endif LOG
			message(Arpa_Info, "Warning: alias database out of date");
		}
	}


	/*
	**  If necessary, load the DBM file.
	**	If running without DBM, load the symbol table.
	*/

	if (init)
	{
#ifdef LOG
		if (LogLevel >= 6)
		{
			extern char *username();

			syslog(LOG_NOTICE, "alias database %srebuilt by %s",
				automatic ? "auto" : "", username());
		}
#endif LOG
		readaliases(aliasfile, TRUE);
	}
# else DBM
	readaliases(aliasfile, init);
# endif DBM
}
/*
**  READALIASES -- read and process the alias file.
**
**	This routine implements the part of initaliases that occurs
**	when we are not going to use the DBM stuff.
**
**	Parameters:
**		aliasfile -- the pathname of the alias file master.
**		init -- if set, initialize the DBM stuff.
**
**	Returns:
**		none.
**
**	Side Effects:
**		Reads aliasfile into the symbol table.
**		Optionally, builds the .dir & .pag files.
*/

static
readaliases(aliasfile, init)
	char *aliasfile;
	bool init;
{
	register char *p;
	char *rhs;
	bool skipping;
	int naliases, bytes, longest;
	FILE *af;
	void (*oldsigint)();
	ADDRESS al, bl;
	register STAB *s;
	char line[BUFSIZ];

	if ((af = fopen(aliasfile, "r")) == NULL)
	{
		if (tTd(27, 1))
			printf("Can't open %s\n", aliasfile);
		errno = 0;
		NoAlias++;
		return;
	}

# ifdef DBM
	/* see if someone else is rebuilding the alias file already */
	if (flock(fileno(af), LOCK_EX | LOCK_NB) < 0 && errno == EWOULDBLOCK)
	{
		/* yes, they are -- wait until done and then return */
		message(Arpa_Info, "Alias file is already being rebuilt");
		if (OpMode != MD_INITALIAS)
		{
			/* wait for other rebuild to complete */
			(void) flock(fileno(af), LOCK_EX);
		}
		(void) fclose(af);
		errno = 0;
		return;
	}
# endif DBM

	/*
	**  If initializing, create the new DBM files.
	*/

	if (init)
	{
		oldsigint = signal(SIGINT, SIG_IGN);
		(void) strcpy(line, aliasfile);
		(void) strcat(line, ".dir");
		if (close(creat(line, DBMMODE)) < 0)
		{
			syserr("cannot make %s", line);
			(void) signal(SIGINT, oldsigint);
			return;
		}
		(void) strcpy(line, aliasfile);
		(void) strcat(line, ".pag");
		if (close(creat(line, DBMMODE)) < 0)
		{
			syserr("cannot make %s", line);
			(void) signal(SIGINT, oldsigint);
			return;
		}
		dbminit(aliasfile);
	}

	/*
	**  Read and interpret lines
	*/

	FileName = aliasfile;
	LineNumber = 0;
	naliases = bytes = longest = 0;
	skipping = FALSE;
	while (fgets(line, sizeof (line), af) != NULL)
	{
		int lhssize, rhssize;

		LineNumber++;
		p = index(line, '\n');
		if (p != NULL)
			*p = '\0';
		switch (line[0])
		{
		  case '#':
		  case '\0':
			skipping = FALSE;
			continue;

		  case ' ':
		  case '\t':
			if (!skipping)
				syserr("Non-continuation line starts with space");
			skipping = TRUE;
			continue;
		}
		skipping = FALSE;

		/*
		**  Process the LHS
		**	Find the final colon, and parse the address.
		**	It should resolve to a local name -- this will
		**	be checked later (we want to optionally do
		**	parsing of the RHS first to maximize error
		**	detection).
		*/

		for (p = line; *p != '\0' && *p != ':' && *p != '\n'; p++)
			continue;
		if (*p++ != ':')
		{
			syserr("missing colon");
			continue;
		}
		if (parseaddr(line, &al, 1, ':') == NULL)
		{
			syserr("illegal alias name");
			continue;
		}
		loweraddr(&al);

		/*
		**  Process the RHS.
		**	'al' is the internal form of the LHS address.
		**	'p' points to the text of the RHS.
		*/

		rhs = p;
		for (;;)
		{
			register char c;

			if (init && CheckAliases)
			{
				/* do parsing & compression of addresses */
				while (*p != '\0')
				{
					extern char *DelimChar;

					while (isspace(*p) || *p == ',')
						p++;
					if (*p == '\0')
						break;
					if (parseaddr(p, &bl, -1, ',') == NULL)
						usrerr("%s... bad address", p);
					p = DelimChar;
				}
			}
			else
			{
				p = &p[strlen(p)];
				if (p[-1] == '\n')
					*--p = '\0';
			}

			/* see if there should be a continuation line */
			c = fgetc(af);
			if (!feof(af))
				(void) ungetc(c, af);
			if (c != ' ' && c != '\t')
				break;

			/* read continuation line */
			if (fgets(p, sizeof line - (p - line), af) == NULL)
				break;
			LineNumber++;
		}
		if (al.q_mailer != LocalMailer)
		{
			syserr("cannot alias non-local names");
			continue;
		}

		/*
		**  Insert alias into symbol table or DBM file
		*/

		lhssize = strlen(al.q_user) + 1;
		rhssize = strlen(rhs) + 1;

# ifdef DBM
		if (init)
		{
			DATUM key, content;

			key.dsize = lhssize;
			key.dptr = al.q_user;
			content.dsize = rhssize;
			content.dptr = rhs;
			store(key, content);
		}
		else
# endif DBM
		{
			s = stab(al.q_user, ST_ALIAS, ST_ENTER);
			s->s_alias = newstr(rhs);
		}

		/* statistics */
		naliases++;
		bytes += lhssize + rhssize;
		if (rhssize > longest)
			longest = rhssize;
	}

# ifdef DBM
	if (init)
	{
		/* add the distinquished alias "@" */
		DATUM key;

		key.dsize = 2;
		key.dptr = "@";
		store(key, key);

		/* restore the old signal */
		(void) signal(SIGINT, oldsigint);
	}
# endif DBM

	/* closing the alias file drops the lock */
	(void) fclose(af);
	CurEnv->e_to = NULL;
	FileName = NULL;
	message(Arpa_Info, "%d aliases, longest %d bytes, %d bytes total",
			naliases, longest, bytes);
# ifdef LOG
	if (LogLevel >= 8)
		syslog(LOG_INFO, "%d aliases, longest %d bytes, %d bytes total",
			naliases, longest, bytes);
# endif LOG
}
/*
**  FORWARD -- Try to forward mail
**
**	This is similar but not identical to aliasing.
**
**	Parameters:
**		user -- the name of the user who's mail we would like
**			to forward to.  It must have been verified --
**			i.e., the q_home field must have been filled
**			in.
**		sendq -- a pointer to the head of the send queue to
**			put this user's aliases in.
**
**	Returns:
**		none.
**
**	Side Effects:
**		New names are added to send queues.
*/

forward(user, sendq)
	ADDRESS *user;
	ADDRESS **sendq;
{
	char buf[60];
	extern bool safefile();

	if (tTd(27, 1))
		printf("forward(%s)\n", user->q_paddr);

	if (user->q_mailer != LocalMailer || bitset(QBADADDR, user->q_flags))
		return;
	if (user->q_home == NULL)
		syserr("forward: no home");

	/* good address -- look for .forward file in home */
	define('z', user->q_home, CurEnv);
	expand("\001z/.forward", buf, &buf[sizeof buf - 1], CurEnv);
	if (!safefile(buf, user->q_uid, S_IREAD))
		return;

	/* we do have an address to forward to -- do it */
	include(buf, "forwarding", user, sendq);
}
